/*   Little John New Generation v0.1.1

     (c) 2000 Yohann Magnien (magnien@ensicaen.ismra.fr)

     This is an attempt to make an easy portable emulator for nes/snes.

     Snes emulation uses the GREAT SNES9X emulator.
 */

/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2000 Gary Henderson (gary@daniver.demon.co.uk) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (Ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */

#ifdef __CYGWIN32__
#include <windows.h>
#define DELETE "echo o|del "
#else
#define DELETE "rm -f "
#endif

#include "ljng.h"


#include <stdlib.h>

/* SDL */
#include "SDL.h"

/* SNES9X headers */
#include "snes9x.h"
#include "memmap.h"
//#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu.h"
#include "display.h"
#include "gfx.h"
#include "soundux.h"
#include "cheats.h"

#include "font.h"

#ifdef TITLE
#undef TITLE
#define TITLE "Little John New Generation 0.1, powered by Snes9X"
#endif


char *rom_filename;
char *rom_filename_bis;
char *snapshot_filename;
Uint32 Last_Ticks;
char *SoundBuffer;
int NewSound;
char IPSok;
SDL_Color *colors;

Uint8 keyssnes[512];


char Gmode;
uint16 AccessFreezeFunc;
char GFX256;

char SpeedUp;
int OldSkipFrame;


char RGBformat;
uint16 *RGBconvert;
uint16 *ShadeBuff;

char quitemu;

const char *S9xGetFilename (const char *ex);
void Yo_Printc(uint16 x,uint16 y,char c,int s);

void OutOfMemory()
{
  printf("Out of memory ... you should at least 16Megs of free memory!\n");
  exit(0);
}

/*
  Initialisation des variables systemes de la SNES
*/

#if !defined(__GLIBC__) || (__GLIBC__ < 2)
const char *basename (const char *f)
{
  const char *p;
  if ((p = strrchr (f, '/')) != NULL || (p = strrchr (f, '\\')) != NULL)
    return (p + 1);

  return (f);
}
#endif


/***************************************
Routines to open and close freeze-game files. STREAM is defined as a
gzFile if ZLIB is defined else its defined as FILE *. The read_only parameter
is set to TRUE when reading a freeze-game file and FALSE when writing a
freeze-game file.
**************************************/

void S9xParseArg (char **argv, int &index, int argc)
{
  if (!(strcasecmp(argv[index],"-256")))
    {
      //8bits graphics mode
      GFX256=1;
      Settings.Transparency = FALSE;
      Settings.SixteenBit = FALSE;
    }
  if (!(strcasecmp(argv[index],"-rgb555")))
    {
      //8bits graphics mode
      RGBformat=1;
    }
}

void S9xExtraUsage ()
{
  printf("[Special SDL options]\n\n");
  printf("-256\t\t\t8bits graphic mode\n\n");
  printf("-hi\t\t\tHigh resolution support(default: disabled)\n\n");
  printf("-tr\t\t\tTransparency support (default: disabled)\n\n");
  printf("[End of SDL options]\n\n");
}

/*********************
If you're going to be using the simple command line parser, when it
encounters an unknown option it calls S9xUsage which is supposed to report
all options the generic parse knows about (I haven't been keeping it up to
date of late). S9xUsage then, in turn calls S9xExtraUsage which you
implement to report any port-specific options available.
***********************/

void S9xGraphicsMode ()
{
  Gmode=1;
}


void S9xTextMode ()
{
  Gmode=0;
}

/*
The SNES debugger calls these routines to switch from a graphics screen
mode used to display the SNES game to a debugger screen used to display
debugger output. If the SNES screen can be displayed at the same time as
a text display, as would be the case when the host system implements a
graphical window system, or you're not going to support the SNES debugger,
then these routines should do nothing.

On the X Window System UNIX/Linux port, these routines do nothing where as
on the MS-DOS port they switch between a graphics screen mode and a text-only
screen mode.
*/

void ClearFRcnt()
{
	  int i,j;

	if (!GFX256)
	  {
		  if (doubleres)
			{
			  for (i=(screen->h)-15;i<(screen->h);i++)
				for (j=0;j<20;j++)
					((uint16 *)(screen->pixels))[(i<<9)+j]=0;
			}
		  else
			{
				for (i=(screen->h)-15;i<(screen->h);i++)
				for (j=0;j<20;j++)
				((uint16 *)(screen->pixels))[(i<<8)+j]=0;
			}
	  }
	  else
	  {
		  if (doubleres)
			{
			  for (i=(screen->h)-15;i<(screen->h);i++)
				for (j=0;j<20;j++)
					((uint8 *)(screen->pixels))[(i<<9)+j]=0;
			}
		  else
			{
				for (i=(screen->h)-15;i<(screen->h);i++)
					for (j=0;j<20;j++)
						((uint8 *)(screen->pixels))[(i<<8)+j]=0;
			}
	  }
}

bool8 S9xInitUpdate ()
{
  //printf("Call to InitUpdate\n");
  return(TRUE);
}

/*****************************
Called just before Snes9x starts to render a SNES screen. The Windows port
uses this call to lock Direct X screen area to allow exclusive access; on
other existing ports its implemented as an empty function.
***********************************/

bool8 S9xDeinitUpdate (int Width, int Height, bool8 sixteen_bit)
{
   //printf("Call to DeinitUpdate with %d %d %d\n",Width,Height,sixteen_bit);
	
  
  int i,j,k,l;
  uint32 a;
  
  if (GFX256)
    {  
      if (pixsize)
	{//dres support, screen is 512x480
	  if (Width==256)
	    {
	      for (i=0;i<Height;i++)
		{
		  for (j=0;j<256;j++)
		    {		
		      ((uint16 *)(screen->pixels))[(i<<9)+j]=((((uint8 *)GFX.Screen)[(i<<9)+j])<<8)|(((uint8 *)GFX.Screen)[(i<<9)+j]);
		    }
		  memcpy((char *)(screen->pixels)+(i<<10)+512,(char *)(screen->pixels)+(i<<10),512);
		}
	    }
	  else
	    {
	      if (Height<=240)
		{
		  for (i=0;i<Height;i++)
		    {
		      memcpy((char *)(screen->pixels)+(i<<10),GFX.Screen+(i<<9),512);
		      memcpy((char *)(screen->pixels)+(i<<10)+512,GFX.Screen+(i<<9),512);
		    }
		}
	      else
		{
		  memcpy((char *)(screen->pixels),GFX.Screen,512*Height);
		}
	    }
	}

      else
	{//lores, screen is 256x240

	  if (Width==256) 
	    {
	      if (Height<=240)
		{
		  for (i=0,j=0;i<Height;i++)
		    {
		      memcpy((char *)(screen->pixels)+(i<<8),GFX.Screen+j,Width); 
		      j+=GFX.Pitch;
		    }
		}
	      else
		{
		  for (i=0,j=0;i<Height;i+=2)
		    {
		      memcpy((char *)(screen->pixels)+((i>>1)<<8),GFX.Screen+j,Width); 
		      j+=GFX.Pitch;
		    }

		}
	    }
	  else
	    {
	      if (Height<=240)
		{
		  /*		  k=0;
				  for (i=0,j=0;i<Height;i++)
				  {
				  memcpy((char *)(screen->pixels)+(i<<9)+k,GFX.Screen+j,Width); 
				  j+=GFX.Pitch;
				  k+=512;
				  }
				  if (IPPU.RenderThisFrame) SDL_UpdateRect(screen,0,0,Width,Height*2); */
		  for (i=0;i<Height;i++)
		    for (j=0;j<256;j++)
		      ((uint8 *)(screen->pixels))[(i<<8)+j]=((uint8 *)(GFX.Screen))[(i<<9)+(j<<1)];
		  
		}
	      else
		{
		  for (i=0,j=0;i<Height;i+=2)
		    {
		      memcpy((char *)(screen->pixels)+((i>>1)<<8),GFX.Screen+j,Width); 
		      j+=GFX.Pitch;
		    }
		}
	    }
	}

      if (NewMSG)
	{
	  if (doubleres)
	    {
	      for (i=0;i<512;i++)
		for (j=16;j<msg_y;j++)
		  ((uint8 *)(screen->pixels))[(j<<9)+i]|=((uint8 *)(ShadeBuff))[((uint8 *)(screen->pixels))[(j<<9)+i]]|((uint8 *)(screen2->pixels))[(j<<9)+i];

	    }
	  else
	    {
	      for (i=0;i<256;i++)
		for (j=16;j<msg_y;j++)
		  ((uint8 *)(screen->pixels))[(j<<8)+i]|=((uint8 *)(ShadeBuff))[((uint8 *)(screen->pixels))[(j<<8)+i]]|((uint8 *)(screen2->pixels))[(j<<8)+i];
	    }

	}
    }
  else
    {
      //dres support, screen is 512x480
      if (doubleres)
	{
	  if (Width==256)
	    {
	      for (i=0;i<Height;i++)
		{
		  for (j=0;j<256;j++)
		    {		
		      ((uint32 *)(screen->pixels))[(i<<9)+j]=((((uint16 *)GFX.Screen)[(i<<9)+j])<<16)|(((uint16 *)GFX.Screen)[(i<<9)+j]);;	       
		    }
		  memcpy((char *)(screen->pixels)+(i<<11)+1024,(char *)(screen->pixels)+(i<<11),1024);
		}
	    }
	  else
	    {
	      if (Height<=240)
		{
		  for (i=0;i<Height;i++)
		    {
		      memcpy((char *)(screen->pixels)+(i<<11),GFX.Screen+(i<<10),1024);
		      memcpy((char *)(screen->pixels)+(i<<11)+1024,GFX.Screen+(i<<10),1024);
		    }
		}
	      else
		{
		  memcpy((char *)(screen->pixels)+16*1024,GFX.Screen,512*Height*2);
		}
	    }
	}

      else
	//lores, screen is 256x240
	{
	  if (Width==256) 
	    {
	      if (Height<=240)
		{
		  for (i=0,j=0;i<Height;i++)
		    {
		      memcpy((char *)(screen->pixels)+(i<<9),GFX.Screen+j,Width<<1); 
		      j+=GFX.Pitch;
		    }
		}
	      else
		{
		  for (i=0,j=0;i<Height;i+=2)
		    {
		      memcpy((char *)(screen->pixels)+((i>>1)<<9),GFX.Screen+j,Width<<1); 
		      j+=GFX.Pitch;
		    }

		}
	    }
	  else
	    {
	      if (Height<=240)
		{
		  /*		  k=0;
				  for (i=0,j=0;i<Height;i++)
				  {
				  memcpy((char *)(screen->pixels)+(i<<10)+k,GFX.Screen+j,Width<<1); 
				  j+=GFX.Pitch;
				  k+=1024;
				  }
				  if (IPPU.RenderThisFrame) SDL_UpdateRect(screen,0,0,Width,Height*2); */
		  for (i=0;i<Height;i++)
		    for (j=0;j<256;j++)
		      ((uint16 *)(screen->pixels))[(i<<8)+j]=((uint16 *)(GFX.Screen))[(i<<9)+(j<<1)];
		  
		}
	      else
		{
		  for (i=0,j=0;i<Height;i+=2)
		    {
		      memcpy((char *)(screen->pixels)+((i>>1)<<9),GFX.Screen+j,Width<<1); 
		      j+=GFX.Pitch;
		    }
		}
	    }
	}

      if (NewMSG)
	{
	  if (doubleres)
	    {
	      for (i=0;i<512;i++)
		for (j=16;j<msg_y;j++)
		  ((uint16 *)(screen->pixels))[(j<<9)+i]=ShadeBuff[((uint16 *)(screen->pixels))[(j<<9)+i]]|((uint16 *)(screen2->pixels))[(j<<9)+i];

	    }
	  else
	    {
	      for (i=0;i<256;i++)
		for (j=16;j<msg_y;j++)
		  ((uint16 *)(screen->pixels))[(j<<8)+i]=ShadeBuff[((uint16 *)(screen->pixels))[(j<<8)+i]]|((uint16 *)(screen2->pixels))[(j<<8)+i];
	    }

	}
    }


  FR++;

  if (frame>=(Settings.PAL?50:60))
    {
      FRcnt=FR;
      FR=0;
      frame=0;
    }
  
  if (ShowFR)
    {
      ClearFRcnt();
      Yo_Printc(0,(screen->h)-15,(FRcnt/10)+48,0);
      Yo_Printc(10,(screen->h)-15,(FRcnt%10)+48,0);
    }

    if ((!GFX256)&&(RGBformat))
        for (a=0;a<(doubleres?512*480:256*240);a++) ((uint16 *)(screen->pixels))[a]=((uint16 *)(RGBconvert))[((uint16 *)(screen->pixels))[a]];

		  

    if (viewGUI)
      SDL_UpdateRect(screen,0,13<<(pixsize),0,0);
    else
      SDL_UpdateRect(screen,0,0,0,0);


  return(TRUE);
}

void S9xDeinitDisplay ()
{
    printf("Call to DeinitDisplay\n");
  free(GFX.Screen);
  free(GFX.SubScreen);
  free(GFX.ZBuffer);
  free(GFX.SubZBuffer);
  free(ShadeBuff);
}

void S9xInitDisplay (int argc, char **argv)
{
  //  printf("Call to InitDisplay\n");

  int w,h,b;
  char chaine[256];
  uint32 i;

  if (doubleres)
    {
      w=512;
      h=480;
      pixsize=1;
    }       
  else
    {
      w=256;
      h=240;
      pixsize=0;
    }

  if (GFX256) 
    {
      b=8;
      bppix=0;
    }
  else 
    {
      b=16;
      bppix=1;
    }

  /* Initialize the display in a wxh 16-bits */
 
  screen = SDL_SetVideoMode(w, h, b, SDL_SWSURFACE|(fullscreen?SDL_FULLSCREEN:0));
  if ( screen == NULL ) {
    fprintf(stderr, "Couldn't set video mode: %s\n",
	    SDL_GetError());
    exit(1);
  }
  
  screen2 = SDL_CreateRGBSurface(SDL_SWSURFACE,w,h,b,31,63,31,0);

  if (!screen2)
    {
      fprintf(stderr, "Couldn't set print buffer for grahics mode: %s\n",
	      SDL_GetError());
      exit(1);
    }
  
  if ((screen->pitch)!=(GFX256?w:w*2))
  {
	  printf("Problem with screen pitch : %d/%d! Exiting ...\n",(GFX256?w:w*2),screen->pitch);
	  exit(0);
  }
 
  GFX.Screen= (uint8 *)malloc(512*480*(GFX256?1:2));
  GFX.SubScreen = (uint8 *)malloc(512*480*(GFX256?1:2));
  GFX.ZBuffer = (uint8 *)malloc(512*480*(GFX256?1:2));
  GFX.SubZBuffer = (uint8 *)malloc(512*480*(GFX256?1:2));
  GFX.Pitch = 512*(GFX256?1:2);

  

  memset(screen2->pixels,0,w*h<<bppix);

  msg_x=0; msg_y=16;
  sprintf(chaine,"Video mode : %dx%d %dbits\n",w,h,screen->format->BitsPerPixel);
  S9xMessage(0,0,chaine);

  ShadeBuff=(uint16 *)malloc(65536*2);

  for (i=0;i<65536;i++) ShadeBuff[i]=((i>>13)<<11)|((((i>>5)&63)>>2)<<5)|((i&31)>>2);

  if (!GFX256)
    {
      RGBconvert=(uint16 *)malloc(65536*2);
      if (!RGBconvert)
	{
	  OutOfMemory();
	  S9xExit();
	}
      printf("Calc\n");
      for (i=0;i<65536;i++) ((uint16 *)(RGBconvert))[i]=((i>>11)<<10)|((((i>>5)&63)>>1)<<5)|(i&31);
      printf("ok\n");
    }
  else 
    colors  = (SDL_Color *)malloc(256*sizeof(SDL_Color)); 

}

/*****************************************************************

Called once a complete SNES screen has been rendered into the GFX.Screen
memory buffer, now is your chance to copy the SNES rendered screen to the
host computer's screen memory. The problem is that you have to cope with
different sized SNES rendered screens. Width is always 256, unless you're
supporting SNES hi-res. screen modes (Settings.SupportHiRes is TRUE), in
which case it can be 256 or 512. The height parameter can be either 224 or
239 if you're only supporting SNES lo-res. screen modes, or 224, 239, 448 or
478 if hi-res. SNES screen modes are being supported.

All current ports support scaling the SNES screen to fill the host system's
screen, the many ports even supports interpolation - blending the colours of
adjacent pixels to help hide the fact they've been scaled - and scan-line
simulation - slightly darkening every other horizontal line.

Don't forget that if you're just placing the SNES image centerally in the
screen then you might need to clear areas of the screen if the SNES image
changes size between calls to S9xDeinitDisplay. The MS-DOS and UNIX ports
currently don't do this which results in junk being left on the screen if
the ROM changes SNES screen modes.

The sixteen_bit is just a copy of the Settings.SixteenBit setting and if
TRUE indicates a 16-bit SNES screen image has been rendered, 8-bit otherwise.
************************/

void Yo_Printc(uint16 x,uint16 y,char c,int s)
{
  int indice=c-32; //index in the font.h array
  int indicex,indicey;
  void *scr=(s?screen2->pixels:screen->pixels);
  if (indice>=96) indice=0;
  uint16 i,j;
  indicey=(indice>>4)*9;
  indicex=(indice&0xf)<<3;


  if (GFX256)
    {
      if (doubleres)
	{
	  for (i=0;i<8;i++)
	    for (j=0;j<9;j++) if (font[j+indicey][i+indicex]=='#') ((uint8 *)(scr))[((y+j)<<9)+i+x]=255;
	}
      else
	{
	  for (i=0;i<8;i++)
	    for (j=0;j<9;j++) if (font[j+indicey][i+indicex]=='#') ((uint8 *)(scr))[((y+j)<<8)+i+x]=255;
	}
    }
  else
    {
      if (doubleres)
	{
	  for (i=0;i<8;i++)
	    for (j=0;j<9;j++) if (font[j+indicey][i+indicex]=='#') ((uint16 *)(scr))[((y+j)<<9)+i+x]=(31<<11)|(63<<5)|31;
	}
      else
	{
	  for (i=0;i<8;i++)
	    for (j=0;j<9;j++) if (font[j+indicey][i+indicex]=='#') ((uint16 *)(scr))[((y+j)<<8)+i+x]=(31<<11)|(63<<5)|31;
	}
    }


}

void S9xMessage (int type, int number, const char *message)
{
  int l;
  int k;
  int re;

  if (screen2)
    {
      l=strlen(message);
      //little hack here
      if (message[l-1]!='\n') re=1;
      else re=0;
      k=0;
      NewMSG=60*(5+l>>4);
      while (k<l)
	{
	  if (message[k]!='\n')
	    {
	      Yo_Printc(msg_x,msg_y,message[k],1);
	      msg_x+=6;
	    }
	  else 
	    {
	      msg_y+=8;
	      msg_x=0;
	    }

	  if ((k==(l-1))&&re)
	    {
	      msg_y+=8;
	      msg_x=0;
	    }

	  if (msg_x>=(doubleres?512-6:256-6))
	    {
	      msg_x=0;
	      msg_y+=8;
	    }
	  if (msg_y>=(doubleres?480-8:240-8))
	    {
	      msg_x=0;
	      msg_y=16;
	      memset((char *)(screen2->pixels),0,(doubleres?512*480<<bppix:256*240<<bppix));
	    }
	  k++;
	}
    }
  else
    {
      printf("%s\n",message);
    }
}

/***********************************************************

I've started work on converting all the old printfs into calls to this
routine.  When Snes9x wants to display an error, information or warning
message, it calls this routine. Check in messages.h for the types and
individual message numbers that Snes9x currently passes as parameters.

The idea is display the message string so the user can see it, but you
choose not to display anything at all, or change the message based on the
message number or message type.

Eventually all debug output will also go via this function, trace information
already does.
***************************************/


void Snes_Sound_CallBack(void *udata, Uint8 *stream, int len)
{
  if ((len!=(buffsize<<(so.sixteen_bit+so.stereo)))&&(!badsoundbuff)) 
    {
      badsoundbuff=1;
      printf("Sound buffer has incorrect size, try the -buff n options!\n");
    }
  S9xMixSamples(stream,len>>Sound16bits);
}




bool8 S9xOpenSoundDevice(int mode, bool8 stereo, int buffer_size)
{
  SDL_AudioSpec wanted,obt[1];

  if (!enablesound) return(FALSE);

  /* Set the audio format */
  wanted.freq = snd_frequency; //mode;
  wanted.format = (Sound16bits?AUDIO_S16:AUDIO_S8);
  wanted.channels = (SoundStereo?2:1);    /* 1 = mono, 2 = stereo */
  
  wanted.samples = buffsize;  /* Good low-latency value for callback */
  wanted.callback = Snes_Sound_CallBack;
  wanted.userdata = NULL;

  /* Open the audio device, forcing the desired format */
  if ( SDL_OpenAudio(&wanted, obt) < 0 ) {
    printf("Couldn't open audio: %s\n", SDL_GetError());
    return(FALSE);
  }

  if (obt)
    {
      printf("Obtained : freq %d format %d channels %d samples %d\n",obt->freq,obt->format,obt->channels,obt->samples);
      buffsize=buffer_size=obt->samples;
      snd_frequency=obt->freq;
      SoundStereo=obt->channels-1;
      Sound16bits=(obt->format==AUDIO_S16);
    }


  printf("Sound Initialized\n");

  so.stereo = SoundStereo;
  so.sixteen_bit = Sound16bits;
  so.playback_rate = snd_frequency;
  so.buffer_size = buffsize;

  if (so.stereo) so.buffer_size*=2;
  if (so.sixteen_bit) so.buffer_size*=2;

  printf("Sound buffer size %d, Max %d\n",so.buffer_size,MAX_BUFFER_SIZE);
  
  so.encoded = FALSE; 

  return(TRUE);
}


/*****************************************************************

S9xInitSound calls this function to actually open the host operating system's
sound device, or initialise the sound card in MS-DOS port. 

The mode parameter is the value passed in on the command line with the -r
command line flag, assuming you're using the Snes9x parser. Its meant to
indicate what playback the sound hardware should be set to, value 1 to 7.
I think the real SNES sound chip playback rate is 30kHz, but such high
playback rates take a lot of native CPU power to emulate. The default
playback rate is 22kHz for the MS-DOS and UNIX ports.

The stereo flag indicates if the user wants stereo sound. Again, stereo
sound takes more CPU to power to emulate compared to mono sound.

The buffer_size value indicates what sample buffer size the user wants,
usually zero, meaning you should pick the value best suited to the current
playback rate. Sound data is normally passed to the sound hardware in
blocks, the smaller the block the less latency between the SNES game playing
a sound and it being heard by the user. But if you pick a too smaller value,
and you're having to periodically poll the operating system to see if it can
accept more sound data, then the sound output will break up because other
actions such as rendering the SNES screen can prevent you from polling the
hardware often enough and the operating system runs out of sound data to
play.

The MS-DOS port uses a buffer size of 128 samples since the sound card
sends an interrupt when more data is required which is acted upon promptly,
where as the Linux and Solaris ports use a buffer size of 512 samples or
more depending on the playback rate. Stereo and 16-bit sound both double the
actual size of the buffer in bytes.
**************************************/

uint32 S9xReadJoypad (int which1_0_to_4)
{
  uint32 val=0x80000000;
  Sint16 pos;

  fillkeys(keyssnes);

  switch (which1_0_to_4)
    {
    case 0:
		if (UseJoystickPl1)
		{
				SDL_JoystickUpdate();
				pos=SDL_JoystickGetAxis(joy1,0);
				if (pos<-8192) val|=SNES_LEFT_MASK;
				if (pos>8192) val|=SNES_RIGHT_MASK;
				pos=SDL_JoystickGetAxis(joy1,1);
				if (pos>8192) val|=SNES_DOWN_MASK;
				if (pos<-8192) val|=SNES_UP_MASK;
				switch (joy1nb)
				{
				case 8:
					if (SDL_JoystickGetButton(joy1,7)) val|=SNES_START_MASK;
				case 7:
					if (SDL_JoystickGetButton(joy1,6)) val|=SNES_SELECT_MASK;
				case 6:
					if (SDL_JoystickGetButton(joy1,5)) val|=SNES_TR_MASK;
				case 5:
					if (SDL_JoystickGetButton(joy1,4)) val|=SNES_TL_MASK;
				case 4:
					if (SDL_JoystickGetButton(joy1,3)) val|=SNES_A_MASK;
				case 3:
					if (SDL_JoystickGetButton(joy1,2)) val|=SNES_B_MASK;
				case 2:
					if (SDL_JoystickGetButton(joy1,1)) val|=SNES_X_MASK;
				case 1:
					if (SDL_JoystickGetButton(joy1,0)) val|=SNES_Y_MASK;
				
				}
		}


			if (keyssnes[ljng_keysym[L_1]]==SDL_PRESSED) val|=SNES_TL_MASK;

			if (keyssnes[ljng_keysym[R_1]]==SDL_PRESSED) val|=SNES_TR_MASK;
			if (keyssnes[ljng_keysym[X_1]]==SDL_PRESSED) val|=SNES_X_MASK;
			if (keyssnes[ljng_keysym[Y_1]]==SDL_PRESSED) val|=SNES_Y_MASK;
			if (keyssnes[ljng_keysym[B_1]]==SDL_PRESSED) val|=SNES_B_MASK;
			if (keyssnes[ljng_keysym[A_1]]==SDL_PRESSED) val|=SNES_A_MASK;
			if (keyssnes[ljng_keysym[START_1]]==SDL_PRESSED) val|=SNES_START_MASK;
			if (keyssnes[ljng_keysym[SELECT_1]]==SDL_PRESSED) val|=SNES_SELECT_MASK;
			if (keyssnes[ljng_keysym[UP_1]]==SDL_PRESSED) val|=SNES_UP_MASK;
	  		if (keyssnes[ljng_keysym[DOWN_1]]==SDL_PRESSED) val|=SNES_DOWN_MASK;
			if (keyssnes[ljng_keysym[LEFT_1]]==SDL_PRESSED) val|=SNES_LEFT_MASK;
			if (keyssnes[ljng_keysym[RIGHT_1]]==SDL_PRESSED) val|=SNES_RIGHT_MASK;
		
		break;
    case 1:
		if (UseJoystickPl2)
		{
				SDL_JoystickUpdate();
				pos=SDL_JoystickGetAxis(joy2,0);
				if (pos<-8192) val|=SNES_LEFT_MASK;
				if (pos>8192) val|=SNES_RIGHT_MASK;
				pos=SDL_JoystickGetAxis(joy2,1);
				if (pos>8192) val|=SNES_DOWN_MASK;
				if (pos<-8192) val|=SNES_UP_MASK;
				switch (joy2nb)
				{
				case 8:
					if (SDL_JoystickGetButton(joy2,7)) val|=SNES_START_MASK;
				case 7:
					if (SDL_JoystickGetButton(joy2,6)) val|=SNES_SELECT_MASK;
				case 6:
					if (SDL_JoystickGetButton(joy2,5)) val|=SNES_TR_MASK;
				case 5:
					if (SDL_JoystickGetButton(joy2,4)) val|=SNES_TL_MASK;
				case 4:
					if (SDL_JoystickGetButton(joy2,3)) val|=SNES_A_MASK;
				case 3:
					if (SDL_JoystickGetButton(joy2,2)) val|=SNES_B_MASK;
				case 2:
					if (SDL_JoystickGetButton(joy2,1)) val|=SNES_X_MASK;
				case 1:
					if (SDL_JoystickGetButton(joy2,0)) val|=SNES_Y_MASK;
				
				}
		}
      if (keyssnes[ljng_keysym[L_2]]==SDL_PRESSED) val|=SNES_TL_MASK;
      if (keyssnes[ljng_keysym[R_2]]==SDL_PRESSED) val|=SNES_TR_MASK;
      if (keyssnes[ljng_keysym[X_2]]==SDL_PRESSED) val|=SNES_X_MASK;
      if (keyssnes[ljng_keysym[Y_2]]==SDL_PRESSED) val|=SNES_Y_MASK;
      if (keyssnes[ljng_keysym[B_2]]==SDL_PRESSED) val|=SNES_B_MASK;
      if (keyssnes[ljng_keysym[A_2]]==SDL_PRESSED) val|=SNES_A_MASK;
      if (keyssnes[ljng_keysym[START_2]]==SDL_PRESSED) val|=SNES_START_MASK;
      if (keyssnes[ljng_keysym[SELECT_2]]==SDL_PRESSED) val|=SNES_SELECT_MASK;
      if (keyssnes[ljng_keysym[UP_2]]==SDL_PRESSED) val|=SNES_UP_MASK;
      if (keyssnes[ljng_keysym[DOWN_2]]==SDL_PRESSED) val|=SNES_DOWN_MASK;
      if (keyssnes[ljng_keysym[LEFT_2]]==SDL_PRESSED) val|=SNES_LEFT_MASK;
      if (keyssnes[ljng_keysym[RIGHT_2]]==SDL_PRESSED) val|=SNES_RIGHT_MASK;
      break; 
    default:
      break;
    }

  return(val);
}

/****************************************

This function is called to return a bit-wise mask of the state of one of the
five emulated SNES controllers. Return 0 if you're not supporting controllers
past a certain number or return the mask representing the current state of
the controller number passed as a parameter or'ed with 0x80000000.

Symbolic constants are defined in snes9x.h indicating the bit positions of
the various SNES buttons and direction indicators; they're all in the form
SNES_X_MASK where X is the SNES controller button name.

The MS-DOS and X Window System ports record what keys are currently pressed
and use that to build up a mask, the Windows port polls the operating system
when S9xReadJoypad is called to find out what keys are pressed. All ports
also implement host joysticks and joy-pads via this interface.
*********************************************/

bool8 S9xReadMousePosition (int which1_0_to_1, int &x, int &y, uint32 &buttons)
{
  printf("Reading mouse pos\n");

  x=0;
  y=0;
  buttons=0;
  return(TRUE);
}
/*******************************************************************************

Used by Snes9x to get the current position of the host pointing device,
usually a mouse, used to emulated the SNES mouse. Snes9x converts the x and
y values to delta values required by the SNES mouse, so the actual x and y
values are unimportant, only the change in value since the last call to
this function is used. 

Graphical windowing systems normally restrict the movement of the pointer on
the screen, if you're porting to such an environment you might want to make
a note of the change in position in the mouse since the last time you asked
the operating system the mouse position, add this change in value to some
saved x and y value, the reposition the pointer back to the centre of the
SNES display window. The saved x and y values will be the values returned
by this function.

The buttons return value is a bit-wise mask of the two SNES mouse buttons,
bit 0 for button 1 (left) and bit 1 for button 2 (right).
********************************************/


bool8 S9xReadSuperScopePosition (int &x, int &y, uint32 &buttons)
{
  printf("Reading superscope pos\n");
  return(FALSE);
}


void S9xSetPalette ()
{
  int i;

  if (!GFX256) return;

  uint16 Brightness = 3; //(IPPU.MaxBrightness) * 140;

  for (i = 0; i < 256; i++)
    {
      colors[i].r = ((PPU.CGDATA[i] >> 0) & 0x1F) << Brightness;
      colors[i].g = ((PPU.CGDATA[i] >> 5) & 0x1F) << Brightness;
      colors[i].b = ((PPU.CGDATA[i] >> 10) & 0x1F) << Brightness;

      //      printf("Color : %d %d %d\n",colors[i].r,colors[i].g,colors[i].b);
    } 

  if (viewGUI)
    {
      colors[32].r=colors[32].g=colors[32].b=255;
      colors[7].r=255; colors[7].g=0; colors[7].b=0;
      colors[23].r=128; colors[23].g=0; colors[23].b=0;
      colors[38].r=64; colors[38].g=0; colors[38].b=0;
    }
  
  if ((ShowFR)||(NewMSG))
    {
      colors[255].r=colors[255].g=colors[255].b=255;
    }
 


   SDL_SetColors(screen, colors, 0, 256);
}

void S9xSyncSpeed ()
{
  int haswait;
  int freq;

  if (LJNG_Process()==0) quitemu=1;

   S9xProcessEvents(FALSE);

  if (NewMSG)
    {
      NewMSG--;
      if (!NewMSG)
	{
	  memset(screen2->pixels,0,(doubleres?512*480<<bppix:256*240<<bppix));
	  msg_x=0; msg_y=16;
	}
    }

  if (Settings.PAL) freq=1000/50;
  else freq=1000/60;

  haswait=0;
  if (!SpeedUp)
    {
      while (SDL_GetTicks()-Last_Ticks<freq) haswait=1;

      Last_Ticks+=freq; //SDL_GetTicks();
    }

  if (Settings.SkipFrames== AUTO_FRAMERATE)
    {
      if (!haswait) 
	{
	  IPPU.RenderThisFrame=FALSE;
	  IPPU.SkippedFrames++;
	}
      else
	{
	  IPPU.RenderThisFrame = TRUE;
	  IPPU.SkippedFrames = 0;
        }
      if (IPPU.SkippedFrames>30) 
	  {
		  IPPU.RenderThisFrame=TRUE;  //do not go beyond 60 frames skipped
		  Last_Ticks=SDL_GetTicks();
		  IPPU.SkippedFrames=0;
	  }

    }
  else
    {
      if (++IPPU.FrameSkip > Settings.SkipFrames)
	{
	  IPPU.RenderThisFrame=TRUE;
	  IPPU.FrameSkip=0;
	}
      else
	{
	  IPPU.RenderThisFrame=FALSE;
	}
    }
  
  frame++;
}

void S9xInitInputDevices()
{
}

void S9xSetTitle(char *title)
{
  SDL_WM_SetCaption(title,title);
}

void InlineHelp()
{
	char chaine[256];
  
	memset(screen2->pixels,0,(pixsize?512*480<<bppix:256*240<<bppix));
	msg_x=0; msg_y=16;

	S9xMessage(0,0,"[Little John New Generation 0.1]\n");
	S9xMessage(0,0,"(c) 2000 Yoyo & Snes9x Team\n");
	
	sprintf(chaine,"[Video] %dx%d %dbits\n",screen->w,screen->h,screen->format->BitsPerPixel);
	S9xMessage(0,0,chaine);
	sprintf(chaine,"[Sound] %dKHz %dbits %s Buffer=%d bytes\n",snd_frequency,(Sound16bits+1)<<3,(SoundStereo?"stereo":"mono"),buffsize);
	S9xMessage(0,0,chaine);
	S9xMessage(0,0,"[Active keys]\n");
	S9xMessage(0,0," F1         -> inline help\n");
	S9xMessage(0,0," F2-F8      -> load state\n");
	S9xMessage(0,0," Shift+F2-F8-> save state\n");
	S9xMessage(0,0," F9         -> toggle fullscreen\n");
	S9xMessage(0,0," F10        -> reset\n");
	S9xMessage(0,0," F11        -> toggle RGB555/RGB565 format\n");
	S9xMessage(0,0," F12        -> show frame rate\n");
	S9xMessage(0,0," TAB        -> fast forward\n");
}

char *GetSnapName()
{
  static char chaine[256];
  char romname[100];
  unsigned char i,j,k,l;

  strcpy(chaine,path_snap);

  i=0;
  j=k=l=255;
  while (CurrentROMFile[i]) 
    {
      if (CurrentROMFile[i]==92) j=i;  //92="\" 
      if (CurrentROMFile[i]==46) k=i;  //46="."
      if (CurrentROMFile[i]=='/') l=i; 
      i++;
    }
  
  CurrentROMFile[k]=0;
  if (l==255) strcpy(romname,&CurrentROMFile[(unsigned char)(j+1)]);  
  else strcpy(romname,&CurrentROMFile[(unsigned char)(l+1)]);

  if (k=255)
    {
      strcat(chaine,romname);
      strcat(chaine,".bmp");
    }
  else
    {
      romname[k+1]=0;
      strcat(chaine,".bmp");
    }

  return(chaine);
}

void S9xProcessEvents (bool8 block)
{
  bool8 Save;

  fillkeys(keyssnes);
  //  printf("Process events\n");
  if (viewGUI) S9xSetPalette();
 
  if (keyssnes[ljng_keysym[QUIT]]==SDL_PRESSED) 
    {
      quitemu=1;
      quit=1;
    }

  if ((keyssnes[ljng_keysym[SHIFTL]]==SDL_PRESSED)||(keyssnes[ljng_keysym[SHIFTR]]==SDL_PRESSED)) Save=1;
  else Save=0;

 
  if (!AccessFreezeFunc)
    {
      if (keyssnes[ljng_keysym[HELP]]==SDL_PRESSED)
	{
	  InlineHelp();
	  AccessFreezeFunc=30;
	}
      if (keyssnes[SDLK_F2]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,0));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,0));
	  AccessFreezeFunc=30;  
	}
      if (keyssnes[SDLK_F3]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,1));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,1));
	  AccessFreezeFunc=30;  
	}
      if (keyssnes[SDLK_F4]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,2));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,2));
	  AccessFreezeFunc=30;  
	}
      if (keyssnes[SDLK_F5]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,3));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,3));
	  AccessFreezeFunc=30;  
	}
      if (keyssnes[SDLK_F6]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,4));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,4));
	  AccessFreezeFunc=30;  
	}
      if (keyssnes[SDLK_F7]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,5));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,5));
	  AccessFreezeFunc=30;  
	}
      if (keyssnes[SDLK_F8]==SDL_PRESSED)
	{
	  if (Save) S9xFreezeGame(S9xChooseFilename (FALSE,6));
	  else S9xUnfreezeGame(S9xChooseFilename (TRUE,6));
	  AccessFreezeFunc=30;  
	}

      if (keyssnes[SDLK_F12]==SDL_PRESSED)
	{
	  if (Save) 
	    {
	      SDL_SaveBMP(screen,GetSnapName()); 
	      AccessFreezeFunc=30; 
	      S9xMessage(0,0,"Saving snapshot\n");
	    }
	  else
	    {
	      ShowFR^=1;
	      AccessFreezeFunc=30;  
	      if (!ShowFR)
		{
		  ClearFRcnt();
		  SDL_UpdateRect(screen,0,0,0,0);
		}
	    }
	}
    
      if (keyssnes[SDLK_F9]==SDL_PRESSED)
	{
	  SDL_WM_ToggleFullScreen(SDL_GetVideoSurface());
	  fullscreen^=1;
	  AccessFreezeFunc=30;  
	}
      
      if (keyssnes[SDLK_F10]==SDL_PRESSED) 
	{
	  S9xReset();
	  AccessFreezeFunc=30;  
	}

      if (keyssnes[SDLK_F11]==SDL_PRESSED) 
	{
	  RGBformat^=1;
	  AccessFreezeFunc=30;  
	}
      
    }
  else
    AccessFreezeFunc--;

  if (keyssnes[ljng_keysym[ACCEL]]==SDL_PRESSED) 
    {
      if (!SpeedUp)
	{
	  SpeedUp=1;
	  OldSkipFrame=Settings.SkipFrames;
	  Settings.SkipFrames=30;
	}
    }
  else 
    {
      if (SpeedUp)
	{
	  SpeedUp=0;
	  Last_Ticks=SDL_GetTicks();
	  Settings.SkipFrames=OldSkipFrame;
	}
    }

  
}

void S9xGenerateSound ()
{
}

void S9xExit ()
{
  char st[256];
  if (enablesound) {SDL_PauseAudio(1); SDL_CloseAudio();}
  S9xSetSoundMute (TRUE);
  S9xDeinitDisplay ();
  Memory.SaveSRAM (S9xGetFilename (".srm"));
  S9xSaveCheatFile (S9xGetFilename (".cht"));
  Memory.Deinit ();
  S9xDeinitAPU ();

#ifdef NETPLAY_SUPPORT
  if (Settings.NetPlay)
    S9xNetPlayDisconnect ();
#endif
    
  if (IPSok)
    {
      strcpy(st,DELETE);
      strcat(st,rom_filename_bis);

      system(st);
    }

  SDL_FreeSurface(screen2);
  SDL_FreeSurface(screen);

    
  exit (0);
}

void _makepath (char *path, const char *, const char *dir,
		const char *fname, const char *ext)
{
  if (dir && *dir)
    {
      strcpy (path, dir);
      strcat (path, SLASH_STR);
    }
  else
    *path = 0;
  strcat (path, fname);
  if (ext && *ext)
    {
      strcat (path, ".");
      strcat (path, ext);
    }
}

void _splitpath (const char *path, char *drive, char *dir, char *fname,
		 char *ext)
{
  *drive = 0;

  char *slash = strrchr (path, '/');
  if (!slash)
    slash = strrchr (path, '\\');

  char *dot = strrchr (path, '.');

  if (dot && slash && dot < slash)
    dot = NULL;

  if (!slash)
    {
      strcpy (dir, "");
      strcpy (fname, path);
      if (dot)
        {
	  *(fname + (dot - path)) = 0;
	  strcpy (ext, dot + 1);
        }
      else
	strcpy (ext, "");
    }
  else
    {
      strcpy (dir, path);
      *(dir + (slash - path)) = 0;
      strcpy (fname, slash + 1);
      if (dot)
	{
	  *(fname + (dot - slash) - 1) = 0;
	  strcpy (ext, dot + 1);
	}
      else
	strcpy (ext, "");
    }
}

void S9xAutoSaveSRAM ()
{
  Memory.SaveSRAM (S9xGetFilename (".srm"));
}

const char *GetHomeDirectory ()
{
  return (getenv ("HOME"));
}

const char *S9xGetDirectory ()
{
  const char *snapshot;
    
  return (path_srm);
}

const char *S9xGetFilename (const char *ex)
{
  static char filename [PATH_MAX];
  char drive [_MAX_DRIVE];
  char dir [_MAX_DIR];
  char fname [_MAX_FNAME];
  char ext [_MAX_EXT];

  _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
  strcpy (filename, S9xGetDirectory ());
  strcat (filename, fname);
  strcat (filename, ex);

  return (filename);
}

const char *S9xGetROMDirectory ()
{
  return (roms_dir);
}

const char *S9xChooseFilename (bool8 read_only,uint8 num)
{
  char def [PATH_MAX];
  char title [PATH_MAX];
  char drive [_MAX_DRIVE];
  char dir [_MAX_DIR];
  char ext [_MAX_EXT];

  char *filename;

  _splitpath (Memory.ROMFilename, drive, dir, def, ext);
  strcat (def, ".sa0");
  def[strlen(def)-1]=num+48;  //set the number

  filename=(char *)malloc(256);
  strcpy(filename,path_save);
  strcat((char *)filename,def);

  S9xMessage(0,0,(read_only?"Loading state\n":"Saving state\n"));

  return (filename);
}

bool8 S9xOpenSnapshotFile (const char *fname, bool8 read_only, STREAM *file)
{
  char filename [PATH_MAX];
  char drive [_MAX_DRIVE];
  char dir [_MAX_DIR];
  char ext [_MAX_EXT];

  //_splitpath (fname, drive, dir, filename, ext);

  strcpy(filename,fname);
    
  /*#ifdef ZLIB
    if (read_only)
    {
    if ((*file = OPEN_STREAM (filename, "rb")))
    return (TRUE);
    }
    else
    {
    if ((*file = OPEN_STREAM (filename, "wb")))
    {
    chown (filename, getuid (), getgid ());
    return (TRUE);
    }
    }
    #else
    char command [PATH_MAX];
    
    if (read_only)
    {
    sprintf (command, "gzip -d <\"%s\"", filename);
    if (*file = popen (command, "r"))
    return (TRUE);
    }
    else
    {
    sprintf (command, "gzip --best >\"%s\"", filename);
    if (*file = popen (command, "wb"))
    return (TRUE);
    }
    #endif*/

  if (read_only)
    {
      if (*file = OPEN_STREAM(filename,"rb")) return(TRUE);
    }
  else
    {
      if (*file = OPEN_STREAM(filename,"wb")) return(TRUE);
    }
  return (FALSE);
}

void S9xCloseSnapshotFile (STREAM file)
{
#ifdef ZLIB
  CLOSE_STREAM (file);
#else
  //pclose (file);
  CLOSE_STREAM (file);
#endif
}

char *TryToApplyIPS(char *filename)
{
  int i,j;
  uint32 Address;
  uint8 Byte;
  uint16 PatchSize,rlen;
  char filenameIPS[256];
  char *temp;
  uint32 filesize;
  FILE *fips,*dest,*src;

  char dir [_MAX_DIR];
  char drive [_MAX_DRIVE];
  static char name [_MAX_FNAME];
  char ext [_MAX_EXT];
  char fname [_MAX_PATH];
  _splitpath (filename, drive, dir, name, ext);
  strcat(name,".");
  strcat(name,ext);

  j=-1;
  i=0;
  while (filename[i]!=0)
    {
      if (filename[i]=='.') j=i;
      i++;
    }

  if (j==-1) return NULL;
  memcpy(filenameIPS,filename,j);
  filenameIPS[j]=0;
  strcat(filenameIPS,".ips");

  fips=fopen(filenameIPS,"rb");
  if (fips==NULL) return NULL;
  S9xMessage(0,0,"IPS file found\n");
  
  fread(filenameIPS,5,1,fips);
  filenameIPS[5]=0;
  if (strcmp(filenameIPS,"PATCH"))
    {
      printf("Invalid IPS file! %s\n",filenameIPS);
    }
  else
    {
      j=0;

      dest=fopen(name,"wb");
      if (!dest) 
	{
	  printf("Error, IPS autopatcher need to create a temporary ROM!\n");
	  return NULL;
	}
      src=fopen(filename,"rb");
      if (!src) 
	{
	  printf("Original ROM file not found!\n");
	  return NULL;
	}
      temp=(char *)malloc(65536);
      if (!temp) OutOfMemory();

      fseek(src,0,SEEK_END);
      filesize=ftell(src);;
      fseek(src,0,SEEK_SET);

      Address=filesize;
      while (filesize>=65536)
	{
	  filesize-=65536;
	  fread(temp,65536,1,src);
	  fwrite(temp,65536,1,dest);
	}
      if (filesize)
	{
	  fread(temp,filesize,1,src);
	  fwrite(temp,filesize,1,dest);
	}
      fclose(src);

      filesize=Address;
      Address=0;
      while (1)
	{
	  fread(filenameIPS,1,3,fips);
	  filenameIPS[3]=0;
	  if (!strcmp(filenameIPS,"EOF")) {printf("Success!\n"); break;}
	  if (feof(fips)) break;
      
	  fseek(fips,-3,SEEK_CUR);

	  fread(&Address,1,3,fips);
          Address=((Address&0xff)<<16)|(Address&0xFF00)|((Address&0xFF0000)>>16);
	  
	  fread(&PatchSize,1,2,fips);
	  PatchSize=((PatchSize&0xff)<<8)|((PatchSize&0xff00)>>8);
	  
	  if (!PatchSize)  //RLE compressed data
	    {
	      fread(&rlen,1,2,fips);
	      rlen=((rlen&0xff)<<8)|((rlen&0xff00)>>8);
	      fread(&Byte,1,1,fips);
	    }

	  if (Address<filesize)
	    {	      
	      fseek(dest,Address,SEEK_SET);

	     
	      if (PatchSize)
		{
		  Address+=PatchSize;
		  if (filesize<Address) 
		    {
		      filesize=Address;
		      //		      printf("Extending\n");
		    }
		  //		  printf("Regular patch %04X %d\n",Address,PatchSize);
		  fread(temp,PatchSize,1,fips);
	       	  fwrite(temp,PatchSize,1,dest);
		}
	      else
		{
		  Address+=rlen;
		  if (filesize<Address) 
		    {
		      filesize=Address;
		      //		       printf("Extending\n");
		    }
		  //		  printf("RLE patcht %04X %d %d\n",Address,rlen,Byte);
		  for (i=0;i<rlen;i++) temp[i]=Byte;
		  fwrite(temp,rlen,1,dest);
		}
	    }
	  else
	    {
	      //	      printf("Extending target from %d to %d\n",filesize,Address);
	      fseek(dest,0,SEEK_END);
	      while (filesize<Address)
		{
		  filesize++;
		  temp[0]=0;
		  fwrite(temp,1,1,dest);
		}
	    }
	
	  //	  printf("File pos %d\n",ftell(fips));
	}
    }
  fclose(fips);
  fclose(dest);
  return name;
}

void Snes9X_Start()
{
  quitemu=0;
  if (enablesound)   SDL_PauseAudio(0);

  while (!quitemu)
    {
      if (!Settings.Paused) S9xMainLoop (); 

      if (Settings.Paused)
	{
	  S9xSetSoundMute (TRUE);
	}

      if (!Settings.Paused)
	{
	  S9xSetSoundMute (FALSE);
	}

    }
}

void Snes_Close_ROM()
{
  char st[256];
  S9xSetSoundMute (TRUE);
  S9xDeinitDisplay ();
  Memory.SaveSRAM (S9xGetFilename (".srm"));
  S9xSaveCheatFile (S9xGetFilename (".cht"));
  Memory.Deinit ();
  S9xDeinitAPU ();
  
  if (IPSok)
    {
      strcpy(st,DELETE);
      strcat(st,rom_filename_bis);

      system(st);
    }

  SDL_FreeSurface(screen2);
  SDL_FreeSurface(screen);
  if (enablesound) {SDL_PauseAudio(1); SDL_CloseAudio();}

  
}



int Init_Run_Snes(int argc,char **argv)
{
  int i;
  /*    if (argc < 2)
	S9xUsage ();*/

  ZeroMemory (&Settings, sizeof (Settings));

  Settings.SoundPlaybackRate = 5; //snd_frequency/1024;
  Settings.Stereo = SoundStereo;
  Settings.SoundBufferSize = 0; //buffsize;
  Settings.SPCTo65c816Ratio = 2;
  Settings.InterpolatedSound = TRUE;
  Settings.DisableSoundEcho = FALSE;
  Settings.APUEnabled = TRUE;
  Settings.H_Max = (100 * SNES_CYCLES_PER_SCANLINE) / 100;
  Settings.SkipFrames = AUTO_FRAMERATE;
  Settings.Shutdown = TRUE;
  Settings.FrameTimePAL = 20000;
  Settings.FrameTimeNTSC = 16667;
  Settings.FrameTime = Settings.FrameTimeNTSC;
  Settings.DisableSampleCaching = FALSE;
  Settings.DisableMasterVolume = FALSE;
  Settings.Mouse = TRUE;
  Settings.SuperScope = TRUE;
  Settings.MultiPlayer5 = TRUE;
  Settings.ControllerOption = SNES_MULTIPLAYER5;
  Settings.Transparency = (GFX256?FALSE:TRUE);
  Settings.SixteenBit = (GFX256?FALSE:TRUE);
  Settings.SupportHiRes = TRUE;
  Settings.NetPlay = FALSE;
  Settings.ServerName [0] = 0;
  Settings.ThreadSound = FALSE;
  Settings.AutoSaveDelay = 30;
  Settings.ForceNoTransparency = (GFX256?TRUE:FALSE);
#ifdef NETPLAY_SUPPORT
  Settings.Port = NETPLAY_DEFAULT_PORT;
#endif

  S9xParseArgs(argv,argc);

  rom_filename = CurrentROMFile;

  Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;

  if (!Memory.Init () || !S9xInitAPU())
    OutOfMemory ();


  S9xInitSound (Settings.SoundPlaybackRate, Settings.Stereo,Settings.SoundBufferSize);
  if (!Settings.APUEnabled)
	
    {
      printf("Sound off\n");
      S9xSetSoundMute (TRUE);
    }


  uint32 saved_flags = CPU.Flags;

#ifdef GFX_MULTI_FORMAT
  S9xSetRenderPixelFormat (RGB565);
#endif

  S9xInitDisplay (argc, argv);

  if (!S9xGraphicsInit ())
    OutOfMemory ();

  if (rom_filename)
    {
      if (rom_filename_bis=TryToApplyIPS(rom_filename)) 
	{
	  IPSok=1;
	}
      else 
	{
	  IPSok=0;
	  rom_filename_bis=rom_filename;
	}

      if (!Memory.LoadROM (rom_filename_bis))
	{
	  char dir [_MAX_DIR];
	  char drive [_MAX_DRIVE];
	  char name [_MAX_FNAME];
	  char ext [_MAX_EXT];
	  char fname [_MAX_PATH];

	  _splitpath (rom_filename, drive, dir, name, ext);
	  _makepath (fname, drive, dir, name, ext);

	  strcpy (fname, S9xGetROMDirectory ());
	  strcat (fname, name);
	  if (ext [0])
	    {
	      strcat (fname, ".");
	      strcat (fname, ext);
	    }
	  _splitpath (fname, drive, dir, name, ext);
	  _makepath (fname, drive, dir, name, ext);
	  if (!Memory.LoadROM (fname))
	    {
	      printf ("Error opening: %s\n", rom_filename);
	      exit (1);
	    }
	}
      Memory.LoadSRAM (S9xGetFilename (".srm"));
      S9xLoadCheatFile (S9xGetFilename (".cht"));

    }
  else
    {
      S9xReset ();
      Settings.Paused |= 2;
    }
  CPU.Flags = saved_flags;


  S9xInitInputDevices ();

  	
  S9xTextMode ();

#ifdef NETPLAY_SUPPORT
  if (strlen (Settings.ServerName) == 0)
    {
      char *server = getenv ("S9XSERVER");
      if (server)
	{
	  strncpy (Settings.ServerName, server, 127);
	  Settings.ServerName [127] = 0;
	}
    }
  char *port = getenv ("S9XPORT");
  if (Settings.Port >= 0 && port)
    Settings.Port = atoi (port);
  else
    if (Settings.Port < 0)
      Settings.Port = -Settings.Port;

  if (Settings.NetPlay)
    {
      int player;

      if (!S9xNetPlayConnectToServer (Settings.ServerName, Settings.Port,
				      Memory.ROMName, player))
	{
	  fprintf (stderr, "\
Failed to connected to Snes9x netplay server \"%s\" on port %d.\n",
		   Settings.ServerName, Settings.Port);
	  S9xExit ();
	}
      fprintf (stderr, "\
Connected to \"%s\" on port %d as player #%d playing \"%s\"\n",
	       Settings.ServerName, Settings.Port, player, Memory.ROMName);
    }
    
#endif

  if (snapshot_filename)
    {
      int Flags = CPU.Flags & (DEBUG_MODE_FLAG | TRACE_FLAG);
      if (!S9xLoadSnapshot (snapshot_filename))
	exit (1);
      CPU.Flags |= Flags;
    }

  S9xGraphicsMode ();

  sprintf (String, "\"%s\" %s: %s", Memory.ROMName, TITLE, VERSION);
  S9xSetTitle (String);
    
  if (!Settings.APUEnabled)
    S9xSetSoundMute (FALSE);

  Last_Ticks=SDL_GetTicks();

  Snes9X_Start();

  Snes_Close_ROM();
  return(0);
}

/*
  Depart du programme
*/

START_EXTERN_C
int snes_main(int argc,char **argv)
{
  snapshot_filename=NULL;

  RGBformat=0;
  AccessFreezeFunc=0;
  Gmode=0;
  GFX256=0;
  rom_filename=rom_filename_bis=snapshot_filename=NULL;
  SpeedUp=0;

  Init_Run_Snes(argc,argv);

  return(0);
}

void S9xReinitsound()
{
  S9xSetSoundMute(TRUE);
  
  S9xOpenSoundDevice(snd_frequency,SoundStereo,buffsize);
  S9xSetSoundMute(FALSE);					
}

END_EXTERN_C
